/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include "block_sorting.h"

/*
 * Coeur de l'algorithme du block sorting :
 * Obtenir une chaine de caractere "triée" et réversible
 */
void blockSort(Tchar* data, int size, int* resultRow, Tchar** result){
  int i ;  
  bufferSize = size ;
  int indices[bufferSize] ;  
  ALLOUER(bufferBlock,bufferSize) ;
  memcpy ( bufferBlock, data, bufferSize ) ;

  //faire la rotation sur data -> chaine de fin
  for(i=0 ;i<bufferSize ;i++)
    data[i] = bufferBlock[(i+bufferSize-1)%bufferSize] ;

  //initialiser un tableau de rotation des caracteres
  initTab(indices,bufferSize) ;

  /*on utilise un tri qui va nous permettre de récuperer la derniere colonne
    de la matrice triée*/
  qsort(indices, bufferSize, sizeof(int), compareIndices) ;
  
  for(i=0 ;i<bufferSize ;i++){
    if(indices[i] == 0)
      *resultRow = i ;  
    (*result)[i] = data[indices[i]] ;
  }
  free(bufferBlock) ;
}


/*
 *Fonction inverse
 */
void blockSortReverse(Tchar* data, int size, int resultRow, Tchar** result){
  int i,j ;
  Tchar* firstCol ;
  ALLOUER(firstCol,size) ;
  memcpy(firstCol,data,size) ;
  qsort(firstCol, size, sizeof(Tchar), compareTChar) ;

  int tabId[size] ;
  fill(tabId,size,0) ;
  int indices[size] ;
  //  fill(indices,size,0) ;
  //pour chacun des caracteres de la premiere colonne
  for(i=0 ;i<size ;i++){
    //pour chaun des caracteres de la derniere colonne
    for(j=0 ;j<size ;j++){
      //trouver l'indice correspondant
      if((int)firstCol[j] == (int)data[i] && !tabId[j])
	break ;
    }
    tabId[j]=1 ;
    indices[i] = j ;
  }

  i = resultRow ;
  for(j = size - 1 ; j >= 0 ; j--){
    (*result)[j] = data[i] ;
    i = indices[i] ;
  }
  free(firstCol) ;
}

/*
 * move to front :
 * Remplace chaque caractere par sa position dans une liste.
 * A chaque nouveau caractere lu celui-ci vient au début du tableau
 * (plus petit indice)
 */
void moveToFront(FILE* in, FILE* out){
  Tchar i, c ;
  int indices[TCHAR_SIZE] ;

  //initialisation du tableau 00 => 0, 01 => 1, ... FF => 255
  initTab(indices,TCHAR_SIZE) ;

  //lire tous les caracteres du fichier
  while(fread(&c,sizeof(Tchar),1,in)){
    //rechercher l'indice courrant du caractere
    i=0 ;
    while(i<TCHAR_SIZE &&  indices[i] != c)
      i++ ;
    //écrire le caractere
    fwrite(&i,sizeof(Tchar),1,out) ;
    //bouger le caractere au début du tableau
    for( ;i ;--i)
      indices[i] = indices[i-1] ;
    indices[0] = c ;

  } 
}

/*
 * move to front inverse
 */
void moveToFrontReverse(FILE* in, FILE* out){
  Tchar i, c ;
  int indices[TCHAR_SIZE] ;

  //initialisation du tableau 00 => 0, 01 => 1, ... FF => 255
  initTab(indices,TCHAR_SIZE) ;

  while(fread(&i,sizeof(Tchar),1,in)){
    c = indices[i] ;
    //écrire le caractere correspondant
    fwrite(&c,sizeof(Tchar),1,out) ;
    
    //bouger le caractere au début du tableau
    for( ;i ;--i)
      indices[i] = indices[i-1] ;
    indices[0] = c ;
  }
}

/*
 * rle de base :
 * remplacer toutes les suites de au moins deux 0 par 0 suivit du nombre de 0
 */
void rle(FILE* in, FILE* out){
  int c, count=0 ;
  while ((c = fgetc (in)) != EOF){
    /*
      lire chacun des caracteres
      tant que l'on a des 0 compter
      sinon, si le compteur n'est pas vide, 
      saisir le nombre de 0 et reinitialiser le compteur
    */
    if(!c){ 
      if(count >= TCHAR_SIZE-1){
	fputc( 0, out ) ;
	fputc( count, out ) ;
	count = 0 ;
      }
      count++ ;
    }else{
      if(count){
	fputc( 0, out ) ;
	fputc( count, out ) ;
	count = 0 ;
      }
      fputc( c, out ) ;
    }
  }
  if(count){
    fputc( 0, out ) ;
    fputc( count, out ) ;
    count = 0 ;
  }
}

/*
 * rle inverse
 */
void rleReverse(FILE* in, FILE* out){
  int c,count ;
  while ((c = fgetc (in)) != EOF){
    if(!c){
      count=fgetc (in) ;
      while(count){
	fputc( 0, out ) ;
	count-- ;
      }
    }else
      fputc( c, out ) ;      
  }
}

/*
 * Autre version de rle sur tous les caracteres :
 * pour chaque suite d'au moins 2 caracteres C semblables,
 * on insere C C N où N est le nombre de caractères de la suite -2
 */
void rle2(FILE* in, FILE* out){
  int c, last=EOF, count=0 ; 
  while ((c = fgetc(in)) != EOF){
    if(c==last && count < TCHAR_SIZE-1){
      count++ ;
    }else{
      if(count){
	count -- ;//-1 pour l'ajout de la ligne suivante
	fputc(last, out) ;
	fputc(count, out) ;
	count = 0 ;
      }
      fputc(c, out) ;
      last=c ;
    }
  }
}

/*
 * inverse de rle2
 */
void rleReverse2(FILE* in, FILE* out){
  int c, last=EOF, count ;
  while ((c = fgetc(in)) != EOF){
    fputc(c, out) ;
    if (c == last){
      count = fgetc(in) ;
      while (count > 0){
	fputc(c, out) ;
	count-- ;
      }
      last = EOF ; 
    }else
      last = c ;        
  }
}

/*
 * Fonction permettant de lancer la compression sur un fichier 'in'
 * vers un autre fichier 'out' passés en paramètre.
 * 
 */
void blockSorting(FILE* in, FILE* out,
		  unsigned int maxBlockSize, int rlea, int matrice){
  if(maxBlockSize>MAX_BUFFER_SIZE)
    maxBlockSize = MAX_BUFFER_SIZE ;


  FILE* tmp_rle2 = tmpfile () ;

#if RLEA>=1
  if(rlea)
    rle2(in, tmp_rle2) ;
  else
    fcopy(in,tmp_rle2) ; 
#else
  fcopy(in,tmp_rle2) ; 
#endif

  rewind(tmp_rle2) ;
  FILE* tmp_mtf = tmpfile () ;
#if BWT>=1
  int blockSize,resultRow ;
  Tchar* block ;
  Tchar* result ;
  ALLOUER(block,maxBlockSize) ;
  ALLOUER(result,maxBlockSize) ;
  while ((blockSize = fread (block,sizeof(Tchar),maxBlockSize,tmp_rle2))){
    if(matrice)
      blockSortMatrice(block,blockSize,&resultRow,&result) ; 
    else
      blockSort(block,blockSize,&resultRow,&result) ; 
#if DEBUG>=2
    fprintf(stderr,"WRITE : blockSize = %d, resultRow = %d\n",
	    blockSize, resultRow) ;
#endif
    fwrite(&blockSize, sizeof(int), 1, tmp_mtf) ;
    fwrite(&resultRow, sizeof(int), 1, tmp_mtf) ;
    fwrite(result, sizeof(Tchar), blockSize, tmp_mtf) ;   
  }
  free(block) ;
  free(result) ;
#else
  fcopy(tmp_rle2,tmp_mtf) ; 
#endif
  fclose(tmp_rle2) ;

  rewind(tmp_mtf) ;
  FILE* tmp_rle = tmpfile () ;
#if MTF>=1
  moveToFront(tmp_mtf,tmp_rle) ;
#else
  fcopy(tmp_mtf,tmp_rle) ;    
#endif
  fclose(tmp_mtf) ;
   
  rewind (tmp_rle) ;  
  FILE* tmp_sf = tmpfile () ;
#if RLE>=1
  rle(tmp_rle, tmp_sf) ;
#else
  fcopy(tmp_rle, tmp_sf) ; 
#endif
  fclose(tmp_rle) ;

  rewind(tmp_sf) ;
#if SF>=1
    struct shannon_fano *sf ;
    struct bitstream *bs ;
    unsigned int size=0 ;
    int i ;
    sf = open_shannon_fano() ;
    bs = open_bitstream_file(out, "w") ;
    while ((i = fgetc (tmp_sf)) >=0){
      size++ ;
      put_entier_shannon_fano(bs,sf,i) ;      
    }
    close_bitstream_file(bs) ;
    close_shannon_fano(sf) ;
    fwrite(&size, sizeof(unsigned int), 1, out) ;        
#else
  fcopy(tmp_sf, out) ;
#endif
  fclose(tmp_sf) ;
}

/*
 * Fonction permettant de lancer la décompression sur un fichier 'in'
 * vers un autre fichier 'out' passés en paramètre.
 * 
 */
void blockSortingReverse(FILE* in, FILE* out, int rlea, int matrice){
 
  FILE* tmp_sf = tmpfile () ;
#if SF>=1
    struct shannon_fano *sf ;
    struct bitstream *bs ;
    unsigned int size=0 ;
    int i ;
    fseek(in, -4, SEEK_END) ;
    fread(&size,sizeof(unsigned int),1,in) ;
    rewind(in) ;

    sf = open_shannon_fano() ;
    bs = open_bitstream_file(in, "r") ;
    for( ;size ;--size){
      i=get_entier_shannon_fano(bs,sf) ;
      fputc( i, tmp_sf ) ;
    }
    close_bitstream_file(bs) ;
    close_shannon_fano(sf) ;
#else
  fcopy(in,tmp_sf) ;
#endif

  rewind(tmp_sf) ;
  FILE* tmp_rle = tmpfile () ;      
#if RLE>=1
  rleReverse(tmp_sf,tmp_rle) ;
#else
  fcopy(tmp_sf,tmp_rle) ;
#endif
  fclose(tmp_sf) ;

  rewind (tmp_rle) ;
  FILE* tmp_mtf = tmpfile () ;
#if MTF>=1
  moveToFrontReverse(tmp_rle,tmp_mtf) ;
#else
  fcopy(tmp_rle,tmp_mtf) ;    
#endif
  fclose(tmp_rle) ;

  rewind (tmp_mtf) ;    
  FILE* tmp_rle2 = tmpfile () ;
#if BWT>=1
  int blockSize,resultRow ;
  Tchar* block ;
  Tchar* result ;
  while(fread(&blockSize,sizeof(int),1,tmp_mtf)){
    fread(&resultRow,sizeof(int),1,tmp_mtf) ;
#if DEBUG>=2
    fprintf(stderr,"READ : blockSize = %d, resultRow = %d\n",
	    blockSize, resultRow) ;
#endif
    //on ne connais pas la taille du block à l'avance 
    //donc allocation à chaque block
    ALLOUER(block,blockSize) ;
    ALLOUER(result,blockSize) ;
    fread (block,sizeof(Tchar),blockSize,tmp_mtf) ;
    if(matrice)
      blockSortReverseMatrice(block,blockSize,resultRow,&result) ;
    else
      blockSortReverse(block,blockSize,resultRow,&result) ;
    fwrite(result,sizeof(Tchar),blockSize,tmp_rle2) ;
    free(block) ;
    free(result) ;
  }
#else
  fcopy(tmp_mtf,tmp_rle2) ;
#endif
  fclose(tmp_mtf) ;

  rewind (tmp_rle2) ;    

#if RLEA>=1
  if(rlea)
    rleReverse2(tmp_rle2,out) ;
  else
    fcopy(tmp_rle2,out) ;
#else
  fcopy(tmp_rle2,out) ;
#endif
  fclose(tmp_rle2) ;

}

/**
 * Utilitaires
 */

/*
 * Rempli un tableau avec une valeur
 */
void fill(int* tab, int size, int value){
  while(size--)
    tab[size]=value ;
}

/*
 * Initialise les valeurs d'un tableau par valeur = indice.
 */
void initTab(int* tab, int size){
  while(size--)
    tab[size]=size ;
}

 
/*
 * Comparer deux chaines
 */
int compareStr (const void *arg1, const void *arg2){ 
  return(strcmp(*(char **)arg1, *(char **)arg2)) ; 
}                                                                               

/*
 * Comparer deux caracteres
 */
int compareTChar (const void *arg1, const void *arg2){ 
  return (*(Tchar*)arg1) - (*(Tchar*)arg2) ; 
}                                                                               

/*
 * Fonction de comparaison permettant de trouver le vecteur de transformation
 * du block avec rotation + tri
 */
int compareIndices(const void *arg1, const void *arg2){
  int i=0, result=0, nb1 = *(int *)arg1, nb2 = *(int *)arg2, c1, c2 ;  
  while(i<bufferSize && result==0){
    c1 = bufferBlock[(nb1 + i) % bufferSize] ;
    c2 = bufferBlock[(nb2 + i) % bufferSize] ;
    result = c1 - c2 ;
    i++ ;
  }
  return result ;
}

/*
 * procédure permettant de copier un fichier
 */
void fcopy(FILE* in, FILE* out){
  assert(in != NULL) ;
  assert(out != NULL) ;
  Tchar buffer[COPY_BUFFER_SIZE] ;
  unsigned int size =0 ;
  while((size = fread(&buffer, sizeof(Tchar), COPY_BUFFER_SIZE, in))>0){
    fwrite(&buffer, sizeof(Tchar), size, out) ;
  }
}

/*
 * Fonction qui retourne la taille d'un fichier ( /!\ :réinitialise la position)
 */
unsigned int filesize(FILE* f){
  assert(f != NULL) ;
  fseek(f, 0L, SEEK_END) ;
  unsigned int size = ftell(f) ;
  fseek(f, 0L, SEEK_SET) ;
  return size ;
}



/**
 * Version lente (matrices)
 */

Tchar** allocation_matrice_carree_char(int nbe){
  Tchar** matrice ;
  ALLOUER(matrice,nbe) ;
  int i ;
  for(i=0 ;i<nbe ;i++)
    ALLOUER(matrice[i],nbe) ;

  return matrice ;

}

void liberation_matrice_carree_char(Tchar **table, int nbe){
  int i ;
  for(i=0 ;i<nbe ;i++)
    free(table[i]) ;  
  free(table) ;
}

void affiche_matrice_carree(int nbe, Tchar **a, FILE* f) {
  int i, j ;
   
  for(j=0 ;j<nbe ;j++){
    for(i=0 ;i<nbe ;i++)
      fprintf(f, " %c", a[j][i]) ;
    fprintf(f, "\n") ;
  }
}

void blockSortMatrice(Tchar* data, int size, int* resultRow, Tchar** result){
  //On crée une matrice carré virtuelle contenant les textes décalés
  int i,j ;
  Tchar** matrice = allocation_matrice_carree_char(size) ;
  for(i=0 ;i<size ;i++){
    for(j=0 ;j<size ;j++){
      matrice[i][j]=data[(j+i)%size] ;
    }
  }
  //On trie les lignes.
  qsort((Tchar *)matrice, size, sizeof(Tchar *), compareStr) ;

  /*On extrait la dernière colonne qui dans les cas réel contient
    de nombreux caractères identiques successif */
  for(i=0 ;i<size ;i++){
    (*result)[i]=matrice[i][size-1] ;
    if(strcmp((char*)matrice[i],(char*)data)==0)
      (*resultRow)=i ;
  }
  liberation_matrice_carree_char(matrice,size) ;
}

void blockSortReverseMatrice(Tchar* data, int size,
			     int resultRow, Tchar** result){
  int i,j,row,col ;
  Tchar** matrice = allocation_matrice_carree_char(size) ;
  Tchar* firstCol ;
  int tabId[size] ;
  ALLOUER(firstCol,size) ;
  //Initialiser la première colonne
  memcpy(firstCol,data,size) ;
  qsort(firstCol, size, sizeof(Tchar), compareTChar) ;
  
  //Trouver les indices d'inversions
  int indices[size] ;
  fill(tabId,size,0) ;
  //  fill(indices,size,0) ;
  for(i=0 ;i<size ;i++){
    for(j=0 ;j<size ;j++){
      if(firstCol[i] == data[j] && !tabId[j])
	break ;
    }
    tabId[j]=1 ;
    indices[i] = j ;
  }

  //Remplir la premiere et derniere colonne
  for(i=0 ;i<size ;i++){
    matrice[i][0]=firstCol[i] ;
    matrice[i][size-1]=data[i] ;
  }

  //Remplir les colonnes intermédiaires
  for(col=1 ;col<size-1 ;col++){  
    for(row=0 ;row<size ;row++){
      matrice[row][col] = matrice[indices[row]][col-1] ;
    }
  }

  //Resultat
  memcpy((*result),matrice[resultRow],size) ;
  liberation_matrice_carree_char(matrice,size) ;
}

