#include "bits.h"
#include "entier.h"

/*
 * Les fonctions de ce fichier permette d'encoder et de dcoder
 * des entiers en utilisant des codes statiques.
 */

/*
 * Codage d'un entier (entre 0 et 32767 inclus) en une chaine de bits
 * qui est crite dans le bitstream.
 *
 * Le nombre est cod par la concatnation du PREFIXE et SUFFIXE
 * Le suffixe est en fait le nombre entier sauf le premier bit a 1
 * 
 * Nombre de bits |    PRFIXE     | nombres cods | SUFFIXE
 *       0        |       00       |      0        |
 *     	 1        |       010      |  1 (pas 0)    |
 *     	 2        |       011      |     2-3       | 2=0 3=1
 *     	 3        |      1000      |     4-7       | 4=00 5=01 6=10 7=11
 *     	 4        |      1001      |     8-15      | 8=000 ... 15=111
 *     	 5        |      1010      |    16-31      | 16=0000 ... 31=1111
 *     	 6        |      1011      |    32-63      |
 *     	 7        |      11000     |    64-127     |
 *     	 8        |      11001     |   128-255     |
 *     	 9        |      11010     |   256-511     |
 *     	 10       |      11011     |   512-1023    |
 *     	 11       |      11100     |  1024-2047    |
 *     	 12       |      11101     |  2048-4097    |
 *     	 13       |      11110     |  4096-8191    |
 *     	 14       |      111110    |  8192-16383   |
 *     	 15       |      111111    | 16384-32767   |
 *
 * Je vous conseille de faire EXIT si l'entier est trop grand.
 *
 */

static char *prefixes[] = { "00", "010", "011", "1000", "1001", "1010", "1011",
			    "11000", "11001", "11010", "11011", "11100",
			    "11101", "11110", "111110", "111111" } ;

void put_entier(struct bitstream *b, unsigned int f)
{

  int nbr= nb_bits_utile(f) ;
  if( nbr >= TAILLE(prefixes)) //
    EXIT ;

  //mise en place du prfixe
  put_bit_string(b , prefixes[nbr]) ;

  //rccupration du suffix
  if( nbr > 1 )
    put_bits(b,nbr-1,f) ;

}

/*
 * Cette fonction fait l'inverse de la prcdente.
 *
 * Une implmentation propre, extensible serait d'utiliser
 * un arbre binaire comme pour le dcodage d'Huffman.
 * Ou bien parcourir l'arbre des tats 8 bits par 8 bits (voir le cours)
 * Mais je ne vous le demande pas
 */

unsigned int get_entier(struct bitstream *b)
{
  int index ;
  char prefixe [6] = "" ; //chaine qui va contenir le prefixe du nombre

  //tant que l'on a pas captur le prfixe prendre le premier bit et comparer aux prfixes
  do{
    strcat(prefixe,!get_bit(b)?"0":"1") ;
    for(index=0 ;index<TAILLE(prefixes) ;index++){
      if(strcmp(prefixe,prefixes[index]) == 0){
	if(index)
	  return pow2(index-1)+get_bits(b,index-1) ;
	else
	  return  0 ;
      }
    }
  }  while(1) ;
}

/*
 * Operation sur des entiers signs
 *
 * Si l'entier est sign, il est prcd d'un bit  1:negatif et 0:positif
 * On considre que l'entier 0 est positif donc on pourra ajouter
 * 1 aux nombres ngatif pour rcuprer la place du zero ngatif.
 *    2 --> 0 2
 *    1 --> 0 1
 *    0 --> 0 0
 *   -1 --> 1 0
 *   -2 --> 1 1
 *   -3 --> 1 2
 *
 */

void put_entier_signe(struct bitstream *b, int i)
{
  put_bits(b,1, i < 0 ) ;
  put_entier(b,i < 0 ? -i-1 : i) ;
}
/*
 *
 */
int get_entier_signe(struct bitstream *b)
{
  int signe = get_bit(b) ;
  return (signe?-1:1) * (get_entier(b) +(signe) ) ;
}
