/**
 *
 * Copyright (C) 2013, 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include <time.h>
#include "bases.h"
#include "block_sorting.h"


void print_help(char* argv[]){
  printf("Usage: %s -i <filename> -o <filename> -[c|x] [options]\n", argv[0]) ;
  printf("Options :\n") ;
  printf(" -b <size>       \
 Indiquer la taile des blocks. Par défaut %d, maximum %d\n",
	 DEFAULT_BLOCK_SIZE,MAX_BUFFER_SIZE) ;
  printf(" -c              \
 Compresser le fichier source vers le fichier de destination.\n") ;
  printf(" -h              \
 Affichier l'aide.\n") ;
  printf(" -i <filename>   \
 Fichier source.\n") ;
  printf(" -m              \
 Utilisation de la méthode lente avec les matrices. \
 A éviter sur les gros fichiers.\n") ;
  printf(" -o <filename>   \
 Fichier destination. \
 Si un fichier ayant le même nom existe déjà il sera remplacé.\n") ; 
  printf(" -r              \
 Activer le RLE sur tous les caractères avant BWT.\n");
  printf(" -x              \
 Décompresser le fichier source vers le fichier de destination.\n") ;
}


int main(int argc, char **argv){
  FILE * input_f ;
  FILE * output_f ;
  int option,action=0, matrice=0, rlea=0;
  unsigned int blockSize=DEFAULT_BLOCK_SIZE ;

  char *input_filename = "" ;
  char *output_filename = "" ;
  extern char *optarg ;
  extern int optind, optopt, opterr ;
  while ((option = getopt(argc, argv, "i:o:xcb:rmh")) != EOF){
    switch (option){
    case 'i':
      input_filename = optarg ;
      break ;
    case 'o':
      output_filename = optarg ;
      break ;
    case 'b':
      if ((sscanf(optarg, "%d", &blockSize) != 1))
	blockSize = MAX_BUFFER_SIZE ;
      break ;
    case 'c':
      action=1 ;
      break ;
    case 'x':
      action=2 ;
      break ;  
    case 'm':
      matrice=1 ;
      break ;
    case 'r':
      rlea=1 ;
      break ;
    case '?': 
    case 'h': 
    default : 
      print_help(argv);
      break ;	
    }
  }

  if(strlen(input_filename) && strlen(output_filename) && action){
#if DEBUG>=1
    clock_t tStart = clock() ;
#endif
    input_f = fopen(input_filename, "r") ;
    assert(input_f != NULL) ;
    output_f = fopen(output_filename, "w") ;
    assert(output_f != NULL) ;

#if DEBUG>=1
    fprintf(stderr,"%s TO %s \n",input_filename,output_filename) ;    
#endif
    if(action==1){
      blockSorting(input_f, output_f,blockSize,rlea,matrice) ;
    }else if(action==2){
      blockSortingReverse(input_f, output_f,rlea,matrice) ;
    }
#if DEBUG>=1
    unsigned int size1 = filesize(input_f) ;
    unsigned int size2 = filesize(output_f) ;
    double ratio = (1-((double)size2/(double)size1))*100 ;
    fprintf(stderr,"==========================\n") ;
    fprintf(stderr,"SOURCE SIZE : %d\n",size1) ;
    fprintf(stderr,"DESTINATION SIZE : %d\n",size2) ;
    fprintf(stderr,"RATIO : %.2f %% \n",ratio) ;
    if(action == 1 && ratio<0)
      fprintf(stderr,"VERY BAD COMPRESSION.\n") ;
#endif
    fclose(input_f) ;
    fclose(output_f) ;
  

#if DEBUG>=1
    printf("CPU TIME USED : %.2fs\n",
	   (double)(clock() - tStart)/CLOCKS_PER_SEC) ;
#endif
  }
  return 0 ;
}

