/*
 * Le but du shannon-fano dynamique est de ne pas transmettre la table
 * des occurrences. Pour ce faire, on ajoute  la table un symbole ESCAPE
 * qui permet l'ajout d'lments  la table.
 * Le dcompresseur sait qu'aprs un vnement ESCAPE il trouvera
 * la valeur (et non le code) d'un vnement  ajouter  la table.
 */


#include "bits.h"
#include "sf.h"

#define VALEUR_ESCAPE 0x7fffffff /* Plus grand entier positif */

struct evenement
{
  int valeur ;
  int nb_occurrences ;
} ;

struct shannon_fano
{
  int nb_evenements ;
  struct evenement evenements[200000] ;
} ;

/*
 * Allocation des la structure et remplissage des champs pour initialiser
 * le tableau des vnements avec l'vnement ESCAPE (avec une occurrence).
 */
struct shannon_fano* open_shannon_fano()
{
  struct shannon_fano* s_shannon_fano;
  ALLOUER(s_shannon_fano,1);

  s_shannon_fano->evenements[0].valeur = VALEUR_ESCAPE;
  s_shannon_fano->evenements[0].nb_occurrences = 1;
  s_shannon_fano->nb_evenements=1;
  
  return s_shannon_fano;
}

/*
 * Fermeture (libration mmoire)
 */
void close_shannon_fano(struct shannon_fano *sf)
{
  free(sf);
}

/*
 * En entre l'vnement (sa valeur, pas son code shannon-fano).
 * En sortie la position de l'vnement dans le tableau "evenements"
 * Si l'vnement n'est pas trouv, on retourne la position
 * de l'vnement ESCAPE.
 */

static int trouve_position(const struct shannon_fano *sf, int evenement)
{

    int index;
    for(index=0;index<sf->nb_evenements;index++)
      if(evenement==sf->evenements[index].valeur)
	return index;
  
    return trouve_position(sf, VALEUR_ESCAPE) ;

}

/*
 * Soit le sous-tableau "evenements[position_min..position_max]"
 * Les bornes sont incluses.
 *
 * LES FORTES OCCURRENCES SONT LES PETITS INDICES DU TABLEAU
 *
 * Il faut trouver la sparation.
 * La sparation sera code comme l'indice le plus fort
 * des fortes occurrences.
 *
 * La sparation minimise la valeur absolue de la diffrence
 * de la somme des occurrences suprieures et infrieures.
 *
 * L'algorithme (trivial) n'est pas facile  trouver, rflchissez bien.
 */
static int trouve_separation(const struct shannon_fano *sf
			     , int position_min
			     , int position_max)
{

    //trouver la diffrence entre min et max des occurences,
    //vaider 
    
    int i, superieur=0, inferieur=0, difference;

    //faire la somme des valeur entre min et max
    for(i = position_min; i <= position_max; i++) {
      superieur+= sf->evenements[i].nb_occurrences;
    }
    difference=superieur;
    //retourner l'indice le plus fort des occurrences
    for(i = position_min; i <= position_max;i++){
      if(abs(superieur-inferieur)<difference){
	difference = abs(superieur-inferieur);
	//incrmenter la borne inferieur et dcrmenter la borne superieur
	inferieur += sf->evenements[i].nb_occurrences;
	superieur -= sf->evenements[i].nb_occurrences;

      }else 
	return i;
    }
  return 0;
}

/*
 * Cette fonction (simplement itrative)
 * utilise "trouve_separation" pour gnrer les bons bit dans "bs"
 * le code de l'vnement "sf->evenements[position]".
 */

static void encode_position(struct bitstream *bs,struct shannon_fano *sf,
			    int position)
{
  //  trouve_separation(sf,min,max);
  //convertir le code de l'evenement pour gnrer les bons bit dans 
    int position_min=0, position_max=sf->nb_evenements-1, separation=0;
    //prendre la nombre d'venement, 

    while(position_max > position_min){      
      //rcuperer la sparation entre l'evenement courrant et le min courrant
      separation = trouve_separation(sf, position_min, position_max);
      //si la sparation entre les position min et max sont superieur ou gale  la position
      //put (position>sparation)?
      if(position <= separation){
	put_bit(bs, 0);
	position_max = separation;
      }else{
	put_bit(bs, 1);
	position_min = separation+1;
      }
    }




}

/*
 * Cette fonction incrmente le nombre d'occurrence de
 * "sf->evenements[position]"
 * Puis elle modifie le tableau pour qu'il reste tri par nombre
 * d'occurrence (un simple change d'vnement suffit)
 *
 * Les faibles indices correspondent aux grand nombres d'occurrences
 */

static void incremente_et_ordonne(struct shannon_fano *sf, int position)
{

    int index;

    struct evenement tmp;
    sf->evenements[position].nb_occurrences ++;
    //parcourir le tableau, depuis la position de l'evenement vers 0

    for(index=position;index>0;index--){
      //si evenements[position-1].nb_occurrences < evenements[position].nb_occurrences intervertir.
      if(sf->evenements[index-1].nb_occurrences < sf->evenements[index].nb_occurrences){      
	tmp = sf->evenements[index-1];
	sf->evenements[index-1] = sf->evenements[index];
	sf->evenements[index] = tmp;
      }else break; //des qu'il n'y a plus d'inversion sortir de la boucle
    }



}

/*
 * Cette fonction trouve la position de l'vnement puis l'encode.
 * Si la position envoye est celle de ESCAPE, elle fait un "put_bits"
 * de "evenement" pour envoyer le code du nouvel l'vnement.
 * Elle termine en appelant "incremente_et_ordonne" pour l'vnement envoy.
 */
void put_entier_shannon_fano(struct bitstream *bs
			     ,struct shannon_fano *sf, int evenement)
{

    //trouver la position de l'evenement
    int position = trouve_position(sf,evenement);
    encode_position(bs,sf,position);
    //si ESCAPE traiter  part
    if(sf->evenements[position].valeur == VALEUR_ESCAPE){
    //    if(position == trouve_position(sf,VALEUR_ESCAPE)){
      put_bits(bs,sizeof(evenement)*NB_BITS,evenement);
      sf->evenements[sf->nb_evenements].nb_occurrences = 1;
      sf->evenements[sf->nb_evenements].valeur = evenement;
      sf->nb_evenements++;
    }
  
    incremente_et_ordonne(sf,position);

}

/*
 * Fonction inverse de "encode_position"
 */
static int decode_position(struct bitstream *bs,struct shannon_fano *sf)
{




    //capturer les bits jusqu' trouver la position de l'evenement

    int position_min=0, position_max = sf->nb_evenements-1, separation;

    while(position_max>position_min){
      
      separation = trouve_separation(sf, position_min, position_max);
      if(get_bit(bs))
	position_min = separation+1;
      else
	position_max = separation;
    }

    return position_min;

}

/*
 * Fonction inverse de "put_entier_shannon_fano"
 *
 * Attention au pige : "incremente_et_ordonne" change le tableau
 * donc l'vnement trouv peut changer de position.
 */
int get_entier_shannon_fano(struct bitstream *bs, struct shannon_fano *sf)
{

    int position, valeur;
    //prendre la position de  l'evenement
    position = decode_position(bs, sf);
    //    printf("=>%d\n",position);
    //s'il s'agit d'escape le traiter  part
    if(sf->evenements[position].valeur == VALEUR_ESCAPE){
      sf->evenements[sf->nb_evenements].nb_occurrences = 1;
      valeur =  get_bits(bs, sizeof(int)*NB_BITS);
      sf->evenements[sf->nb_evenements].valeur = valeur;     
      sf->nb_evenements++;
    }else
      valeur = sf->evenements[position].valeur;
    
    //faire rordonner  le tableau  la fin
    incremente_et_ordonne(sf, position);
      

    return valeur;


}

/*
 * Fonctions pour les tests, NE PAS MODIFIER, NE PAS UTILISER.
 */
int sf_get_nb_evenements(struct shannon_fano *sf)
{
  return sf->nb_evenements ;
}
void sf_get_evenement(struct shannon_fano *sf, int i, int *valeur, int *nb_occ)
{
  *valeur = sf->evenements[i].valeur ;
  *nb_occ = sf->evenements[i].nb_occurrences ;
}
int sf_table_ok(const struct shannon_fano *sf)
{
  int i, escape ;

  escape = 0 ;
  for(i=0;i<sf->nb_evenements;i++)
    {
      if ( i != 0
	   && sf->evenements[i-1].nb_occurrences<sf->evenements[i].nb_occurrences)
	{
	  fprintf(stderr, "La table des vnements n'est pas trie\n") ;
	  return(0) ;
	}
      if ( sf->evenements[i].valeur == VALEUR_ESCAPE )
	escape = 1 ;
    }
  if ( escape == 0 )
    {
      fprintf(stderr, "Pas de ESCAPE dans la table !\n") ;
      return(0) ;
    }
  return 1 ;
}
